#include <iostream>
#include "HJM.h"
#include "Yield.h"

using namespace std; 			//introduces namespace std

/*construction de l'EDP */
HJMTree::HJMTree(double e, double s, double rr, double lr, double rV, double lV, double r, double V, double t, YieldCurve &ayc)
 : yc(ayc) {
  end = e;
  sigma = s;
  right_r = rr;
  left_r = lr;
  right_V = rV;
  left_V = lV;
  Nr = r;
  NV = V;
  Nt = t;
  
  dr = (right_r - left_r) / Nr;
  dV = (right_V - left_V) / NV;
  dt = end / Nt;

  pos = (yc.r0(0) - left_r) / dr;
  pop = ((pos + 1) * dr + left_r - yc.r0(0)) / dr;
  
  aux = new double* [NV + 1];
  for (int i = 0; i <= NV; i++) {
    aux[i] = new double [Nr + 1];
  }

  upV_aux = dt * sigma * sigma / dV;
  up_down_r_aux = dt * sigma * sigma / (2 * dr * dr);
  
}

/* destruction de l'EDP */
HJMTree::~HJMTree() {
  for (int i = 0; i <= NV; i++) {
    delete [] aux[i];
  }
  delete [] aux;
}

/* creation d'un slice pour l'EDP */
Slice HJMTree::getSlice(double t) {
  Slice *sl = new Slice(Nr, NV, t);
  return *sl;	
}

/* retro-propagation d'un pas */
void HJMTree::pas(Slice &sl) {
  
  for (int i = 0; i < NV; i++) {
    const double V = i * dV + left_V;
    for (int j = 1; j < Nr; j++) {
      double upV, upr, downr, stay_there;
      const double r = j * dr + left_r;

      upV = upV_aux * r;		// upV
      upr = up_down_r_aux * r;		// upr1
      downr = upr;			// downr
      const double temp = yc.r0t(sl.t) + V;
      if (temp > 0) {
        upr += dt * temp / dr;		// upr2
      } else {
        downr -= dt * temp / dr;
      }
      stay_there = 1 - upV - upr - downr;	// stay_there
      
      if (upV < 0 || upr < 0 || downr < 0 || stay_there < 0)
        throw range_error("This is not a probability!");
      aux[i][j] = sl.u[i]     [j]     * stay_there + 
                  sl.u[i + 1] [j]     * upV        + 
                  sl.u[i]     [j + 1] * upr        + 
                  sl.u[i]     [j - 1] * downr      ;
    }
    aux[i][0] = sl.u[i][0];
    aux[i][Nr] = sl.u[i][Nr];
  }
  for (int j = 0; j <= Nr; j++) {
    aux[NV][j] = sl.u[NV][j];
  }  
  sl.t -= dt;
  double **uv = sl.u;
  sl.u = aux;
  aux = uv;
}

/* retro-propagation et actualisation jusqu'a la date t */
void HJMTree::PDE(Slice &sl, double t) {
  cout << sl.t << " -> " << t << endl;
  while (sl.t > t + dt / 10) {
    pas(sl);
    actualisation(sl);
  }
}

/* valeur a l'origine du slice */
double HJMTree::valueAtOrigin(Slice &sl) {
  return pop * sl.u[0][pos] + (1 - pop) * sl.u[0][pos + 1];
}

/* information sur les parametres de discretisation et
   sur les probas de transition */
void HJMTree::print_data() {
  cout << "stay_there: " << 1 - dt * sigma * sigma * left_r / dV - dt * sigma * sigma * left_r / (dr * dr) - dt * left_V / dr;
  cout << "\t" << 1 - dt * sigma * sigma * right_r / dV - dt * sigma * sigma * right_r / (dr * dr) - dt * right_V / dr << endl;

  cout << "upV:        " << dt * sigma * sigma * right_r / dV;
  cout << "\t" << dt * sigma * sigma * left_r / dV << endl;

  cout << "upr:        " << dt * sigma * sigma * right_r / (2 * dr * dr) + dt * right_V / dr;
  cout << "\t" << dt * sigma * sigma * left_r / (2 * dr * dr) + dt * left_V / dr << endl;
  
  cout << "downr:      " << dt * sigma * sigma * right_r / (2 * dr * dr);
  cout << "\t" << dt * sigma * sigma * left_r / (2 * dr * dr) << endl;
  
  cout << "r:          " << yc.r0(0) - Nt * dr << "\t" << yc.r0(0) + Nt * dr << endl;
  cout << "V:          " << 0 << "\t" << Nt * dV << endl;
  
  cout << "dr: " << dr << endl;
  cout << "dV: " << dV << endl;
  cout << "dt: " << dt << endl << endl;

}

/* actualisation d'un slice (si sl.asset est vrai) */
void HJMTree::actualisation(Slice &sl) {
  if (sl.asset) {
    for (int i = 0; i <= NV; i++) {
      for (int j = 0; j <= Nr; j++) {
        sl.u[i][j] /= 1 + (j * dr + left_r) * dt;
      }
    }
  }
}

/* initialisation d'un slice a la date at */
Slice::Slice(int r, int V, double at) {
  Nr = r;
  NV = V;
  t = at;
  asset = true;
  u = new double* [NV + 1];

  for (int i = 0; i <= NV; i++) {
    u[i] = new double [Nr + 1];
    memset(u[i], (Nr + 1) * sizeof(double), 0);
  }
}

/* copie d'un slice */
Slice::Slice(Slice &sl) {
  Nr = sl.Nr;
  NV = sl.NV;
  t = sl.t;
  u = new double* [NV + 1];

  for (int i = 0; i <= NV; i++) {
    u[i] = new double [Nr + 1];
    memcpy(u[i], sl.u[i], (Nr + 1) * sizeof(double));
  }
}

/* destruction d'un slice */
Slice::~Slice() {
  for (int i = 0; i <= NV; i++) {
    delete [] u[i];
  }
  delete [] u;
}

/* valeur maximale entre this slice et sl */
void Slice::maxSlice(Slice &sl) {
  for (int i = 0; i <= NV; i++) {
    for (int j = 0; j <= Nr; j++) {
      u[i][j] = max(u[i][j], sl.u[i][j]);
    }
  }
}

/* valeu maximale entre this slice et d */
void Slice::maxSlice(double d) {
  for (int i = 0; i <= NV; i++) {
    for (int j = 0; j <= Nr; j++) {
      u[i][j] = max(u[i][j], d);
    }
  }
}

/* somme de this slice et d */
void Slice::addSlice(double d) {
  for (int i = 0; i <= NV; i++) {
    for (int j = 0; j <= Nr; j++) {
      u[i][j] += d;
    }
  }
}