/* projet DEA Probabilits et Finances
   cours Monte Carlo
   Andrea Odetti
   Remy Ripoll */

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include "yield.h"
#include "HJM.h"

using namespace std; 			//introduces namespace std

/* ======= parametres de la swaption ========== */
double tenor = 0.5;
int numOfPeriods = 8;
int skip = 0;
double strike = 0.07;

double end;

/* ======= parametres du modele =============== */
double sigma = 0.10;

/* ======= parametres de la discretisation ==== */
double right_r = 0.50, left_r = 0.000;
double right_V = 0.20, left_V = 0.000;

int Nr = 300;
int NV = 10;
int Nt = 1000;

double dr;
double dV;

/* ============================================ */

int estimation = 0;

/* ============================================ */

/* fonction qui lit les parametres du fichier hello.cfg */
void read_parameters() {
  try {
    ifstream input("bermuda.cfg");
    if (input) {
      while(!input.eof()) {
        string s;
        input >> s;
        if (s == "tenor") input >> tenor;
        if (s == "numOfPeriods") input >> numOfPeriods;
        if (s == "skip") input >> skip;
        if (s == "strike") input >> strike;
        if (s == "sigma") input >> sigma;
        if (s == "right_r") input >> right_r;
        if (s == "left_r") input >> left_r;
        if (s == "right_V") input >> right_V;
        if (s == "left_V") input >> left_V;
        if (s == "Nr") input >> Nr;
        if (s == "NV") input >> NV;
        if (s == "Nt") input >> Nt;
        if (s == "estimation") input >> estimation;
      }
    }
  } catch ( ... ) {
    cout << "Error in cfg file!\n";
  }
  end = tenor * numOfPeriods;
  dr = (right_r - left_r) / Nr;
  dV = (right_V - left_V) / NV;
}

/* ============================================ */

void go() {

  cout.setf(ios_base::fixed, ios_base::floatfield);
  cout.precision(8);

  YieldCurve yc;
   
  const double irs = yc.IRS(0, tenor * skip, numOfPeriods - skip, tenor, yc.r0(0), 0);
/* pour un strike a la monnaie il suffit de choisir -1 */
  if (strike < 0) strike = irs;
  
  HJMTree hjm(end, sigma, right_r, left_r, right_V, left_V, Nr, NV, Nt * numOfPeriods, yc);
  hjm.print_data();

/* la bermuda et le swap valent 0 a l'echeance */
  Slice swap = hjm.getSlice(tenor * numOfPeriods);
  Slice bermuda = hjm.getSlice(tenor * numOfPeriods);

/* slice pour calculer la proba que le processus ne touche pas
   les frontieres du domain avant la date d'echeance */
  Slice barriere = hjm.getSlice(tenor * numOfPeriods);
  barriere.asset = false;
  for (int i = 0; i < NV; i++) {
    for (int j = 1; j < Nr; j++) {
      barriere.u[i][j] = 1;
    }
  }

  for (int p = numOfPeriods - 1; p >= skip; p--) {
    const double t = p * tenor;

/* on rajoute le coupon et on retro propage bermuda et swap */
    switch (estimation) {

      case 0: {
        swap.addSlice( - strike * tenor - 1);
      
        hjm.PDE(swap, t);
        hjm.PDE(bermuda, t);

        swap.addSlice(1);
        break;
      }

      case 1: {
        swap.addSlice( - strike * tenor);

        hjm.PDE(swap, t);
        hjm.PDE(bermuda, t);

        for (int i = 0; i <= NV; i++) {
          for (int j = 0; j <= Nr; j++) {
            swap.u[i][j] += 1 - yc.bond(t, t + tenor, left_r + j * dr, left_V + i * dV);
          }
        }
        break;
      }

      case 2: {
        hjm.PDE(bermuda, t);

        for (int i = 0; i <= NV; i++) {
          for (int j = 0; j <= Nr; j++) {
            swap.u[i][j] = yc.swap(t, t, numOfPeriods - p, tenor, strike, left_r + j * dr, left_V + i * dV);
          }
        }
        swap.t = t;
        break;
      }
      
    }
/* la bermuda vaut le max entre sa valeur et le swap */
    bermuda.maxSlice(swap);

  }

/* code pour une swaption europeenne */
  Slice swaption(swap);
  swaption.maxSlice(0);

/* retro propagation jusqu'a la date 0 */
  hjm.PDE(swap, 0);
  hjm.PDE(swaption, 0);
  hjm.PDE(bermuda, 0);
  hjm.PDE(barriere, 0);

  const double val = hjm.valueAtOrigin(swap);
  const double ber = hjm.valueAtOrigin(bermuda);
  const double swn = hjm.valueAtOrigin(swaption);
  const double pbr = hjm.valueAtOrigin(barriere);

  cout << "Valeur de la bermuda:  " << setw(12) << ber << endl;
  cout << "Valeur de la swaption: " << setw(12) << swn << endl;
  cout << "Valeur du swap:        " << setw(12) << val << endl;
  const double fwd = yc.swap(0, tenor * skip, numOfPeriods - skip, tenor, strike, yc.r0(0), 0);
  cout << "Valeur du swap (ff):   " << setw(12) << fwd << endl;
  cout << "Taux de swap:          " << setw(12) << irs * 100 << "%" << endl;
  cout << "Barriere probabilites: " << setw(12) << pbr << endl;
  
  ofstream outf("hjm.txt", ios_base::app);
  outf.setf(ios_base::fixed, ios_base::floatfield);
  outf.precision(10);
  outf << Nr << "\t" << NV << "\t" << Nt << "\t";
  outf << ber << "\t" << swn << "\t" << val << "\t" << pbr << endl;
  outf.close();
}

int main() {
  try {
    read_parameters();
    go();
  } catch (const exception &e) {
    cout << "Got an exception: " << e.what() << endl;
  }
  
  return 0;
}